/* src/outils.c - Divers outils
 *
 * Copyright (C) 2002-2005 David Cortier  <Cesar@irc.jeux.fr>
 *                         Romain Bignon  <Progs@kouak.org>
 *                         Benjamin Beret <kouak@kouak.org>
 *
 * site web: http://sf.net/projects/scoderz/
 *
 * Services pour serveur IRC. Support sur IrcProgs et IRCoderz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * $Id: outils.c,v 1.172 2005/05/06 22:56:10 romexzf Exp $
 */

#include <ctype.h>
#include "main.h"
#include "debug.h"
#include "cs_cmds.h"
#include "admin_manage.h"
#include "hash.h"
#include "config.h"
#include "timers.h"
#include "outils.h"
#include <netinet/in.h>
#include <arpa/inet.h>

static const struct Modes {
	char mode;
	int value;
} UMode[] = {
	{ 'i', N_INV },
	{ 'x', N_CRYPT },
	{ 'r', N_REG },
	{ 'o', N_OPER },
	{ 'k', N_SERVICE },
	{ 'a', N_ADM },
	{ 'p', N_GOD },
	{ 'f', N_FEMME },
	{ 'h', N_HOMME },
	{ 'g', N_DEBUG },
	{ 'w', N_WALLOPS },
	{ 'D', N_DIE },
	{ 'd', N_DEAF },
	{ 'H', N_SPOOF },
},
	CMode[] = {
	{ 'n', C_MMSG },
	{ 't', C_MTOPIC },
	{ 'i', C_MINV },
	{ 'l', C_MLIMIT },
	{ 'k', C_MKEY },
	{ 's', C_MSECRET },
	{ 'p', C_MPRIVATE },
	{ 'm', C_MMODERATE },
	{ 'c', C_MNOCTRL },
	{ 'C', C_MNOCTCP },
	{ 'O', C_MOPERONLY },
	{ 'r', C_MUSERONLY },
};

int chan_check(const char *chan, aNick *nick)
{
	aDNR *dnr;
	int i = 1;

	if(*chan != '#') return csntc(nick, GetReply(nick, L_CHANNELNAME));

	for(;chan[i] && i < REGCHANLEN;++i)
		if(chan[i] == ',' || !isascii(chan[i])) /* bad char or comma */
			return csntc(nick, GetReply(nick, L_USER_INVALID), chan);
	/* if chan[i] != \0 then i > REGCHANLEN */
	if(chan[i])	return csntc(nick, GetReply(nick, L_CHANMAXLEN), REGCHANLEN);

	if((dnr = IsBadChan(chan)))
	{
		if(IsAnAdmin(nick->user))
			csntc(nick, "DNRMask: \2%s\2 Auteur: \2%s\2 Pos le:\2 %s\2 Raison: %s",
				dnr->mask, dnr->from, get_time(nick, dnr->date), dnr->raison);
		else return csntc(nick, GetReply(nick, L_CANREGDNR), chan, dnr->raison);
	}

	return 1;
}

int is_num(const char *num)
{
   while(*num) if(!isdigit(*num++)) return 0;
   return 1;
}

int is_ip(const char *ip)
{
	int dot = 0, digits = 0;
	char d[4] = {0};

	for(;*ip;ip++) {
		if(*ip != '.') {
			if(!isdigit(*ip) || digits == 3) return 0;
			d[digits++] = *ip;
		}
		else {
		    d[digits] = 0;
			if(!digits || ++dot > 3 || atoi(d) > 255) return 0;
			digits = 0;
		}
	}

	d[digits] = 0;
	return (dot == 3 && digits && atoi(d) < 256);
}

void strupr(char *buf)
{
	for(;*buf;++buf) *buf = toupper((unsigned char) *buf);
}

void strlwr(char *buf)
{
	for(;*buf;++buf) *buf = tolower((unsigned char) *buf);
}

void strip_newline(char *string)
{
   register char *p = string;
   while(*p && *p != '\n' && *p != '\r') ++p;
   *p = '\0';
}

char *GetIP(const char *base64)
{
	struct in_addr addr2;

	addr2.s_addr = htonl(base64toint(base64));
	return inet_ntoa(addr2);
}

/*
 * getoption:
 * recherche dans un tableau char ** (de 'parc' entres)
 * l'expression "opt"  partir de l'lment 'deb'
 * -si exp_int est > 0:
 *  cela veut dire qu'on attend une valeur numerique apres le char 'opt'
 *  si c'est le cas, la fct renvoit la valeur, sinon 0
 * -si exp_int est < 0:
 *  cela veut dire que l'on recherche uniquement si 'opt' est dans parv,
 *  si c'est le cas, retourne 1
 * -si exp_int est = 0:
 *  cela veut dire que l'on attend une string apres le 'opt',
 *  si c'est le cas retourne l'index de cette string dans parv, sinon 0.
 *
 */

int getoption(const char *opt, char **parv, int parc, int deb, int exp_int)
{
	for(;deb <= parc;++deb)
		if(!strcasecmp(opt, parv[deb]))
		{
			if(exp_int < 0) return 1;
			if(++deb <= parc)
			{
				if(!exp_int) return deb;
				else if(is_num(parv[deb])) return atoi(parv[deb]);
			}
			break;
		}
	return 0;
}

int split_buf(char *buf, char **parv, int size)
{
	int parc = 0, i;

	while(*buf && parc < size)
	{
		while(*buf == ' ') *buf++ = 0; /* on supprime les espaces pour le dcoupage */
		if(*buf == ':') /* last param */
		{
			parv[parc++] = buf + 1;
			break;
		}
		if(!*buf) break;
		parv[parc++] = buf;
		while(*buf && *buf != ' ') ++buf; /* on laisse passer l'arg.. */
	}
	for(i = parc;i < size;++i) parv[i] = NULL; /* NULLify following items */
	return parc;
}

void parv2msgn(int parc, char **parv, int base, char *buf, int size)
{
	int i = base, toksize, written = 0;

	for(;i <= parc;++i)
	{
		toksize = strlen(parv[i]);

		if(i > base) buf[written++] = ' ';

		if(written + toksize >= size-1)
		{
			Strncpy(buf + written, parv[i], size - written - 1);
			written = size - 1;
			break;
		}
		else
		{
			strcpy(buf + written, parv[i]);
			written += toksize;
		}
	}
	buf[written] = 0;
	return;
}

char *parv2msg(int parc, char **parv, int base, int size)
{
	int i = base, toksize, written = 0;
	static char buf[512];

	if(size > sizeof buf) size = sizeof buf;

	for(;i <= parc;++i)
	{
		toksize = strlen(parv[i]);

		if(i > base) buf[written++] = ' ';

		if(written + toksize >= size-1)
		{
			Strncpy(buf + written, parv[i], size - written - 1);
			written = size - 1;
			break;
		}
		else
		{
			strcpy(buf + written, parv[i]);
			written += toksize;
		}
	}
	buf[written] = 0;
	return buf;
}

char *duration(int s)
{
	static char dur[200];
	int i = 1;
	dur[0] = '\002';

	if(s >= 86400)
		i += sprintf(dur + i, "%d", s/86400), s %= 86400, strcpy(dur + i, "\2 jours \002"), i += 9;
	if(s >= 3600)
		i += sprintf(dur + i, "%d", s/3600), s %= 3600, strcpy(dur + i, "\2 heures \002"), i += 10;
	if(s >= 60)
		i += sprintf(dur + i, "%d", s/60), s %= 60, strcpy(dur + i, "\2 minutes \002"), i += 11;
	if(s) i += sprintf(dur + i, "%d",s), strcpy(dur + i, "\2 secondes");
	else dur[i-2]= 0;

	return dur;
}

time_t convert_duration(const char *p) /* extrait une dure en secondes */
{										/* d'un format XdXhXm.. - Cesar */
	const char *p1 = p;
	time_t result = 0;

	while(*p1)
	{
		if(!isdigit(*p1) && *p1 != 'h' && *p1 != 'H' && *p1 != 'j'
			&& *p1 != 'J' && *p1 != 'd' && *p1 != 'D' && *p1 != 'Y'
			&& *p1 != 'y' && *p1 != 'm')
		{
			result = -1;
			break;
		}
		switch(*p1)
		{
			case 'H': case 'h':
						result += 3600 * strtol(p, NULL, 10);
						p = p1 + 1;
						break;
			case 'm':
						result += 60 * strtol(p, NULL, 10);
						p = p1 + 1;
						break;
			case 'Y': case 'y':
						result += 31536000 * strtol(p, NULL, 10);
						p = p1 + 1;
						break;
			case 'd': case 'j':
			case 'J': case 'D':
						result += 86400 * strtol(p, NULL, 10);
						p = p1 + 1;
						break;
		}
		p1++;
	}
	if(result >= 0) result += strtol(p, NULL, 10);
	return result;
}

char *get_time(aNick *nick, time_t mytime)
{
	static char buftime[40];
	register struct tm *lt = localtime(&mytime);
	const char *ptr = nick ? GetReply(nick, lt->tm_wday + L_DAYS) :
							 DefaultLang->msg[lt->tm_wday + L_DAYS];
	int i = 0;

	while((buftime[i++] = *ptr++));

	buftime[i-1] = ',';
	buftime[i] = ' ';

	snprintf(buftime + i + 1, sizeof buftime, "%d-%02d-%02d %02d:%02d:%02d",
		1900 + lt->tm_year,	lt->tm_mon + 1, lt->tm_mday, lt->tm_hour, lt->tm_min, lt->tm_sec);

	return buftime;
}

char *GetNUHbyNick(aNick *nick, int type)
{
	static char nuhbuf[NUHLEN];
	register char *ptr = nick->nick, *ptr2 = nuhbuf;

	while((*ptr2++ = *ptr++));
	for(*(ptr2-1) = '!', ptr = nick->ident;(*ptr2++ = *ptr++););
	for(*(ptr2-1) = '@',
#ifdef HAVE_CRYPTHOST
		ptr = (type == 1) ? nick->crypt : nick->host
#else
		ptr = nick->host
#endif
	;(*ptr2++ = *ptr++););

	return nuhbuf;
}

char *GetPrefix(aNick *nick)
{
	static char pref[12];
	int i = 0;
	pref[0] = '\0';

	if(IsOper(nick)) strcpy(pref, "\2\0037*\2\003"), i = 6;
	if(nick->user) strcpy(pref + i, IsAdmin(nick->user) ? "\00312^\003" : "\0035!\003");

	return pref;
}

char *GetChanPrefix(aNick *nick, aJoin *join)
{
	static char pref[14];
	int i = 0;

	if(IsOper(nick)) strcpy(pref, "\2\0037*\2\003"), i = 6;
	if(nick->user)
		strcpy(pref + i, IsAdmin(nick->user) ? "\00312^\003" : "\00305!\003"), i += 5;
	if(IsOp(join)) pref[i++] = '@';
	if(IsVoice(join)) pref[i++] = '+';
	pref[i] = 0;
	return pref;
}

char *GetUserOptions(anUser *user)
{
	static char option[120];
	int i = 0;
	option[0] = '\0';

	if(UNopurge(user)) strcpy(option, "NoPurge "), i += 8;
	if(UNoMemo(user)) strcpy(option + i, "NoMemo "), i += 7;
	if(UFirst(user)) strcpy(option + i, "Registering "), i += 12;
	if(UOubli(user)) strcpy(option + i, "Oubli "), i += 6;
	if(UWantX(user)) strcpy(option + i, "VirtualHost "), i += 12;
	if(user->flag & U_ALREADYCHANGE) strcpy(option + i, "UserChanged "), i += 12;
	if(CantRegChan(user)) strcpy(option + i, "CantRegChan "), i += 12;
#ifdef USE_NICKSERV
	if(UPKill(user)) strcpy(option + i, "Protection(Kill) "), i += 17;
	else if(UPNick(user)) strcpy(option + i, "Protection(ChNick) "), i += 19;
#endif
	if(UPReject(user)) strcpy(option + i, "Accs(Rejet) "), i += 13;
	else if(UPAsk(user)) strcpy(option + i, "Accs(Ask) "), i += 11;
	else if(UPAccept(user)) strcpy(option + i, "Accs(Accepte) "), i += 15;
	strcpy(option + i, user->lang->langue);

	if(option[0] == '\0') strcpy(option, "Aucune");
	return option;
}

char *GetAccessOptions(anAccess *acces)
{
	static char options[100];
	int i = 0;
	options[0] = '\0';

	if(AOp(acces)) strcpy(options, " autoop"), i += 7;
	if(AProtect(acces)) strcpy(options + i, " protg"), i += 8;
	if(AVoice(acces)) strcpy(options + i, " autovoice"), i += 10;
	if(ASuspend(acces)) strcpy(options + i, " suspendu");
	if(options[0] == '\0') strcpy(options, " Aucune");
	return options;
}

char *GetChanOptions(aChan *chan)
{
	static char coptions[300];
	int i = 0;
	coptions[0] = 0;

	if(CLockTopic(chan)) strcpy(coptions, " locktopic"), i += 10;
	if(CNoOps(chan)) strcpy(coptions + i, " noops"), i += 6;
	if(CNoBans(chan)) strcpy(coptions + i, " nobans"), i += 7;
	if(CStrictOp(chan)) strcpy(coptions + i, " strictop"), i += 9;
	if(CSetWelcome(chan)) strcpy(coptions + i, " welcome"), i += 8;
	if(CAutoInvite(chan)) strcpy(coptions + i, " autoinvite"), i += 11;
	if(chan->flag & C_ALREADYRENAME) strcpy(coptions + i, " renomm"), i += 8;
	if(CFLimit(chan)) sprintf(coptions + i, " auto-limit(%d,%d)", chan->limit_inc, chan->limit_min);
	if(coptions[0] == 0) strcpy(coptions, " Aucune");

	return coptions;
}

void floating_limit_update_timer(aChan *chan)
{
	if(!CFLimit(chan) && chan->timer) timer_remove(chan->timer), chan->timer = NULL;
	else if(CFLimit(chan) && !chan->timer)
		chan->timer = timer_add(cf_limit_update, TIMER_PERIODIC, callback_fl_update, chan, NULL);
}

void putlog(const char *fichier, const char *fmt, ...)
{
	va_list vl;
	FILE *fp = fopen(fichier, "a");

	if(!fp) return;

	fputc('[', fp);
	fputs(get_time(NULL, CurrentTS), fp);
	fputs("] ", fp);

	va_start(vl, fmt);
	vfprintf(fp, fmt, vl);
	va_end(vl);

	fputc('\n', fp);
	fclose(fp);
}

int check_protect(aNick *from, aNick *cible, aChan *chan)
{
	anAccess *a1, *a2;

	if(from == cible) return 1;
	if(cible->flag & (N_GOD|N_SERVICE))
		return csntc(from, "L'user \2%s\2 est protg", cible->nick);
	/* cible non auth OU from admin et cible non, ou admin de level inf */
	if(!cible->user || from->user->level > cible->user->level) return 1;
	/* from admin de niveau suprieur */
	if(IsAdmin(from->user)) return csntc(from, "L'user \2%s\2 est un Admin "
						"de niveau suprieur ou gal au votre.", cible->nick);
	if(IsAdmin(cible->user)) /* cible est admin */
		return csntc(from, "L'user \2%s\2 est un Administrateur des Services.", cible->nick);
	/* optimisation: on check dabord l'acces de l'execut! */
	if(!(a2 = GetAccessIbyUserI(cible->user, chan)) || ASuspend(a2) || !AProtect(a2)) return 1;
	/* on boucle les access de from pour voir si on en trouve 1 sur le chan */
	if(!(a1 = GetAccessIbyUserI(from->user, chan)) || ASuspend(a1)) return 0;
	/* Should not happen: from is neither Admin nor has access */

	if(a1->level > a2->level) return 1;
	/* on verifie si from a un level superieur a cible, ou si cible n'est pas protect */
	csntc(from, "L'user \2%s\2 est protg ou a un access suprieur au votre sur \2%s\2.",
		cible->nick, chan->chan);
	return 0; /* arrive pour tous les autres cas */
}

int parse_umode(int flag, const char *p)
{	/* default is '+' */
	unsigned int w = 1, i;

	if(!p) return flag;
	for(;*p;p++)
	{
		if(*p == '+') w = 1;
		else if(*p == '-') w = 0;
		else
			for (i = 0;i < ASIZE(UMode);i++)
				if(*p == UMode[i].mode)
				{
					if(w) flag |= UMode[i].value;
					else flag &= ~UMode[i].value;
					break;
				}
	}
	return flag;
}

char *GetModes(int flag)
{
	static char mode[ASIZE(UMode)];
	unsigned int i = 0, j = 0;
	for(;i < ASIZE(UMode);i++) if(flag & UMode[i].value) mode[j++] = UMode[i].mode;
	mode[j] = '\0';
	return mode;
}

int cmodetoflag(int flag, const char *mode)
{
	unsigned int w = 1, i;

	if(!mode) return flag;
	for(;*mode;mode++)
	{
		if(*mode == '+') w = 1;
		else if(*mode == '-') w = 0;
		else for (i = 0;i < ASIZE(CMode);i++)
				if(*mode == CMode[i].mode)
				{
					if(w) flag |= CMode[i].value;
					else flag &= ~CMode[i].value;
					break;
				}
	}
	return flag;
}

int string2scmode(struct cmode *mode, const char *modes, const char *key, const char *limit)
{
    int newmode = cmodetoflag(mode->modes, modes);

	if(newmode & C_MLIMIT)
	{
		int lim = 0;
		if(limit && (lim = strtol(limit, NULL, 10))) mode->limit = lim;
		else if(!(mode->modes & C_MLIMIT)) newmode &= ~C_MLIMIT;/* limit actually invalid */
	}
	else mode->limit = 0;

	if(newmode & C_MKEY)
	{
		if(key && *key) Strncpy(mode->key, key, KEYLEN);
		else if(!*mode->key) newmode &= ~C_MKEY;/* actually no key set */
	} /* -k autoris seulement si key fournie */
	else if(mode->modes & C_MKEY && (!key || !*key)) newmode |= C_MKEY;
	else *mode->key = 0;
	return (mode->modes = newmode);
}

char *GetCModes(struct cmode modes)
{
	static char mode[ASIZE(CMode) + KEYLEN + 20];
	unsigned int i = 0, j = 0;
	for(;i < ASIZE(CMode);i++) if(modes.modes & CMode[i].value) mode[j++] = CMode[i].mode;

	if(modes.modes & C_MLIMIT) j += sprintf(mode + j, " %d", modes.limit);
	if(modes.modes & C_MKEY)
	{
		mode[j++] = ' ';
		strcpy(mode + j, modes.key);
	}
	else mode[j] = '\0';
	return mode;
}

int IsValidNick(char *p)
{
	if(*p == '-' || isdigit(*p)) return 0;
	while(*p)
	{
		if(!isalnum(*p) && !strchr(SPECIAL_CHAR, *p)) return 0;
		p++;
	}
	return 1;
}

int IsValidMail(char *p)
{
	int fa = 0, fd = 0, ispoint = 0;
	if(strchr("-_.@+", *p)) return 0;
	while(*p)
	{
		if(*p == '@')
		{
			if(!fa) fa = 1;
			else return 0;
		}
		if(fa && *p != '.') ispoint = 0;
		if(fa && *p == '.')
		{
			if(ispoint) return 0; /* last char was a dot too.. */
			fd = 1;
			ispoint = 1;
		}
		if(!isalnum(*p) && *p != '.' && *p != '-'
			&& *p != '_' && *p != '@' && (fa || *p != '+')) return 0;
		++p;
	}
	if(fa && fd) return 1;
	return 0;
}

char *RealCmd(const char *cmd)
{	/* returns actual name of a command, given its core name */
	aHashCmd *cmdp = FindCoreCommand(cmd);
	return cmdp ? cmdp->name : "";
}

anUser *ParseNickOrUser(aNick *nick, char *arg)
{
	anUser *user = NULL;

	if(*arg == '%') /* recherche sur nick actif! */
	{
		aNick *n;
		if(*++arg == 0)
			csntc(nick, "Veuillez spcifier un nick aprs '%%' : COMMANDE %%nick-connect [args..]");
		else if(!(n = getnickbynick(arg))) csntc(nick, GetReply(nick, L_NOSUCHNICK), arg);
		else if(!n->user) csntc(nick, GetReply(nick, L_NOTLOGUED), arg);
		else return n->user;
	}
	else if(!(user = getuserinfo(arg))) csntc(nick, GetReply(nick, L_NOSUCHUSER), arg);
	return user;
}

char *str_dup(char **to, const char *from)
{
	if(from && *from)
	{
		*to = realloc(*to, strlen(from) + 1);
		if(!*to) Debug(W_MAX, "EXIT: strdup: no memory for 'to'(%p)=%s 'from'(%p)=%s\n",
			(void *) *to, *to, (const void *) from, from);
		else strcpy(*to, from);
	}
	return *to;
}

char *Strncpy(char *to, const char *from, size_t n)/* copie from dans to. ne copie que n char */
{							/* MAIS AJOUTE LE CHAR \0  la fin, from DOIT donc faire n+1 chars. */
	const char *end = to + n;
	char *save = to;
	while(to < end && (*to++ = *from++));
	*to = 0;
	return save;
}

char *Strtok(char **save, char *str, int sep)/* fonction tire d'ircu, simplifie */
{
	register char *pos = *save; 	/* keep last position across calls */
	char *tmp;

	if(str) pos = str; 	/* new string scan */

	while(pos && *pos && sep == *pos) pos++; /* skip leading separators */

	if (!pos || !*pos) return (pos = *save = NULL); /* string contains only sep's */

	tmp = pos; /* now, keep position of the token */

	if((pos = strchr(pos, sep))) *pos++ = 0;/* get sep -Cesar*/

	*save = pos;
	return tmp;
}

int switch_option(aNick *nick, const char *choix, const char *opt,
	const char *targ, int *flagto, int flag)
{
	if(!choix) *flagto ^= flag;
	else if(!strcasecmp(choix, "on")) *flagto |= flag;
	else if(!strcasecmp(choix, "off")) *flagto &= ~flag;
	else return csntc(nick, GetReply(nick, L_SHOULDBEONOFF));

	if(*flagto & flag) csntc(nick, GetReply(nick, L_OPTIONNOWON), opt, targ);
	else csntc(nick, GetReply(nick, L_OPTIONNOWOFF), opt, targ);
	return 0;
}

int cswall(const char *format, ...)
{
	int i = 0;
	char buf[400];
	va_list vl;

	va_start(vl, format);
	vsnprintf(buf, sizeof buf, format, vl);
	va_end(vl);

	for(;i < adminmax;++i) if(adminlist[i]) csntc(adminlist[i], "%s", buf);

	return 1;
}

inline int fastfmtv(char *buf, const char *fmt, va_list vl)
{
	const char *s = buf;
	while(*fmt)
	{
		if(*fmt == '$')
		{
			register char *tmp = va_arg(vl, char *);
			if(tmp) while(*tmp) *buf++ = *tmp++;
		}
		else *buf++ = *fmt;
		++fmt;
	}
	*buf = 0;
	return buf - s;
}

int fastfmt(char *buf, const char *fmt, ...)
{
	va_list vl;
	int i;

	va_start(vl, fmt);
	i = fastfmtv(buf, fmt, vl);
	va_end(vl);
	return i;
}

int item_isinlist(const int *list, const int max, const int item)
{
	int i = 0;

	for(;i < max;++i) if(list[i] == item) return i;
	return -1;
}

int item_parselist(char *list, int *notes, const int size)
{
    char *ptr = NULL, *p;
    int i = 0;

    for(p = Strtok(&ptr, list, LISTSEP);p && i < size; p = Strtok(&ptr, NULL, LISTSEP))
        if((notes[i] = atoi(p)) > 0 && item_isinlist(notes, i, notes[i]) == -1) i++;

    return i;
}
